//Joris Lambrecht
//20131112
//Oculus Rift Helper file
//  - Sets up stereoscopic camera views with projection translation
//  - Applies barrel distortion and chromatic aberration post-processing (pixel shader)
//  - Connects to Oculus HW to obtain HW parameters and get head tracking data
//requires shaders: "pp_BarrelOR.fx" and "pp_BarrelChromaOR.fx" 
//requires: Oculus.dll in acknex_plugins folder
//
//See Oculus SDK Overview for more information (SDK Version 0.2.5)
//
//Revisions:
//none

#ifndef Oculus_c
#define Oculus_c

#define RAD2DEG  57.2958 
#define METERS_TO_QUANTS 100

//#defines for material skills
#define OculusK1 					skill1 //vecSkill1
#define OculusK2   				skill2
#define OculusK3 					skill3
#define OculusK4			   	skill4
#define OculusX 					skill5 //vecSkill5
#define OculusY   				skill6
#define OculusXCenterOffset 	skill7
#define OculusScale			   skill8
#define OculusChromAb1 			skill9 //vecSkill9
#define OculusChromAb2   		skill10
#define OculusChromAb3 			skill11
#define OculusChromAb4		  	skill12

typedef struct OCULUSPROPS {
	float EyeToScreenDistance;
	float LensSeparationDistance;
	float InterpupillaryDistance;
	float DistortionK[4];
	float DistortionChromAb[4];
	float DistortionScale; //not updated by DLL
	float HScreenSize;
	float VScreenSize;
	unsigned int	HResolution;
	unsigned int	VResolution;
} OCULUSPROPS;

//DLL FUNCTION PROTOTYPES
//Oculus.dll must be in acknex_plugins folder
function readRift(float *rollOR,float *pitchOR,float *yawOR,float *xOR,float *yOR,float *zOR, float dt);
function initRift(OCULUSPROPS *op);


//Global Variables
OCULUSPROPS OculusProps;
var ipi=.064*METERS_TO_QUANTS; //in cm (default value)
var h=0.152; //Horizontal lens center offset (default value)
var player_height = 1.7*METERS_TO_QUANTS; //to base of neck




function translateProjCenter_left(){
	//Note: the Porjection Matrix (matProj) in GS appears to be the transpose of 
	//P specified in the Oculus Rift SDK Manual.  
	//Therefore, in GS, P_ = PH instead of P_ = HP
	float temp[16];
	mat_identity(temp);
	temp[12] = h;
	mat_multiply(matProj, temp);
	mat_set(temp, matProj);
	return 0;
}

function translateProjCenter_right(){
	//Note: the Porjection Matrix (matProj) in GS appears to be the transpose of 
	//P specified in the Oculus Rift SDK Manual.  
	//Therefore, in GS, P_ = PH instead of P_ = HP
	float temp[16];
	mat_identity(temp);
	temp[12] = -h;
	mat_multiply(matProj, temp);
	return 0;
}

MATERIAL* mtl_projAdjust_left =
{
	//The event is called every frame before rendering but after determining the 
	//projection Matrix (based on arc, and clip_near/clip_far, allowing the projection 
	//matrix to be modified.  We need to translate the projection matrix (pg 24)
	event = translateProjCenter_left;
	
	//set default material settings for View
	ambient_blue  = 200;
	ambient_green = 200;
	ambient_red   = 200;
	diffuse_blue  = 200;
	diffuse_green = 200;
	diffuse_red   = 200;
	specular_blue  = 0;
	specular_green = 0;
	specular_red   = 0;
	emissive_blue  = 0;
	emissive_green = 0;
	emissive_red   = 0;
	power = 0;
	albedo = 0;
	
	flags = ENABLE_VIEW;

}

MATERIAL* mtl_projAdjust_right =
{
	//The event is called every frame before rendering but after determining the 
	//projection Matrix (based on arc, and clip_near/clip_far, allowing the projection 
	//matrix to be modified.  We need to translate the projection matrix (pg 24)
	event = translateProjCenter_right;
	
	//set default material settings for View
	ambient_blue  = 200;
	ambient_green = 200;
	ambient_red   = 200;
	diffuse_blue  = 200;
	diffuse_green = 200;
	diffuse_red   = 200;
	specular_blue  = 0;
	specular_green = 0;
	specular_red   = 0;
	emissive_blue  = 0;
	emissive_green = 0;
	emissive_red   = 0;
	power = 0;
	albedo = 0;
	albedo = 0;
	
	flags = ENABLE_VIEW;

}

MATERIAL* mtl_fx_left =  { effect = "pp_BarrelChromaOR.fx"; }
MATERIAL* mtl_fx_right = {	effect = "pp_BarrelChromaOR.fx"; }


VIEW* camera_left_effect =
{ 		
	material = mtl_fx_left;
  	flags = PROCESS_TARGET;
}

VIEW* camera_right_effect =
{ 		
	material = mtl_fx_right;
  	flags = PROCESS_TARGET;
}

VIEW* camera_left =
{ 		
  layer = 0;
  pos_x = 0;
  pos_y = 0;
  stage = camera_left_effect;		  //the stage will apply the distortion pixel shader
  material = mtl_projAdjust_left;  //this will adjust the projection
  //clip_near = 1;
  //clip_far = 1000;
  arc = 78;
  ambient = 0;
  flags = SHOW;
}

VIEW* camera_right =
{ 		
  layer = 0;
  pos_y = 0;
  stage = camera_right_effect;     //the stage will apply the distortion pixel shader
  material = mtl_projAdjust_right; //this will adjust the projection
  //clip_near = 1;
  //clip_far = 1000;
  arc = 78;
  flags = SHOW;
}

function showPropsOR()
{
	DEBUG_VAR(OculusProps.EyeToScreenDistance,30);   //in m
	DEBUG_VAR(OculusProps.LensSeparationDistance,50);//in m
	DEBUG_VAR(OculusProps.InterpupillaryDistance,70);//in m
	DEBUG_VAR(OculusProps.DistortionK[0], 110);			
	DEBUG_VAR(OculusProps.DistortionK[1], 130);
	DEBUG_VAR(OculusProps.DistortionK[2], 150);
	DEBUG_VAR(OculusProps.DistortionK[3], 170);
	DEBUG_VAR(OculusProps.DistortionChromAb[0], 210);			
	DEBUG_VAR(OculusProps.DistortionChromAb[1], 230);
	DEBUG_VAR(OculusProps.DistortionChromAb[2], 250);
	DEBUG_VAR(OculusProps.DistortionChromAb[3], 270);
	DEBUG_VAR(OculusProps.DistortionScale, 310); 
	DEBUG_VAR(OculusProps.HScreenSize, 350);	//in m
	DEBUG_VAR(OculusProps.VScreenSize, 370);	//in m
	DEBUG_VAR(OculusProps.HResolution, 410);
	DEBUG_VAR(OculusProps.VResolution, 430);
}

function updateFOV()
{
	float yfov = 2*atan(OculusProps.VScreenSize*OculusProps.DistortionScale/(2*OculusProps.EyeToScreenDistance))*RAD2DEG;
	float ar = (float)OculusProps.HResolution/(2*(float)OculusProps.VResolution);
		
	camera_left.arc = camera_right.arc = yfov*ar;
}

function getHeadTrackOR(ANGLE* angOR){
	float ORroll, ORpitch, ORyaw, ORx, ORy, ORz;
	readRift(&ORroll,ORpitch,ORyaw,ORx,ORy,ORz,0.02);
	
	angOR.pan = ORyaw * RAD2DEG;
	angOR.tilt = ORpitch * RAD2DEG;
	angOR.roll = -ORroll * RAD2DEG;

}

function updateCamerasOR()
{
	ANGLE  headTrackOR;
	ANGLE  tempang;
	VECTOR tempvec;
	
	getHeadTrackOR(&headTrackOR);
	
	vec_set(tempang, vector(90, 0, 0));
	ang_rotate(tempang, headTrackOR);
		
    // Minimal head modelling.
	vec_set(tempvec, vector(15, ipi/2, 9));
  	vec_rotate(tempvec, tempang);
  	tempvec.z+=player_height;
  	vec_set(camera_left.x, tempvec);
  	
  	vec_set(tempvec, vector(15, -ipi/2, 9));
  	vec_rotate(tempvec, tempang);
  	tempvec.z+=player_height;
  	vec_set(camera_right.x, tempvec);  
  	
  	vec_set(camera_left.pan, tempang);
	vec_set(camera_right.pan, tempang);
	
	
} 

function updateViewsOR(){
	//Both right camera and left camera VIEWS need to be updated whenever the screen size changes
	//Also, the camera effect VIEWS need to be updated to match size and position.
	camera_left.pos_x = 0;
	camera_left_effect.pos_x = 0;
	camera_right.pos_x = screen_size.x/2;
	camera_right_effect.pos_x = screen_size.x/2;
	camera_left.size_x = camera_right.size_x = screen_size.x/2;
	camera_left_effect.size_x = camera_right_effect.size_x = screen_size.x/2;
	camera_left.size_y = camera_right.size_y = screen_size.y;
}


function initOR(){
	initRift(OculusProps);

	reset(camera, SHOW); //disable default camera VIEW
	light_view = NULL;
	
	ipi = OculusProps.InterpupillaryDistance*METERS_TO_QUANTS; 
	h = 1.0-2.0*OculusProps.LensSeparationDistance/OculusProps.HScreenSize;
	
	OculusProps.DistortionScale = 1.5; //this should eventually be set programmatically
	
	//material properties need to be set using floatv conversion (store float value in var)
	mtl_fx_left.OculusX 				  = floatv(0.0); mtl_fx_right.OculusX 			     = floatv(0.5);
	mtl_fx_left.OculusXCenterOffset = floatv(h);	  mtl_fx_right.OculusXCenterOffset = floatv(-h);
	
	mtl_fx_left.OculusY  	= mtl_fx_right.OculusY 		= floatv(0);
	mtl_fx_left.OculusScale = mtl_fx_right.OculusScale = floatv(OculusProps.DistortionScale);
	mtl_fx_left.OculusK1 	= mtl_fx_right.OculusK1 	= floatv(OculusProps.DistortionK[0]); 
	mtl_fx_left.OculusK2 	= mtl_fx_right.OculusK2 	= floatv(OculusProps.DistortionK[1]); 
	mtl_fx_left.OculusK3 	= mtl_fx_right.OculusK3 	= floatv(OculusProps.DistortionK[2]); 
	mtl_fx_left.OculusK4 	= mtl_fx_right.OculusK4 	= floatv(OculusProps.DistortionK[3]);
	mtl_fx_left.OculusChromAb1 	= mtl_fx_right.OculusChromAb1 	= floatv(OculusProps.DistortionChromAb[0]); 
	mtl_fx_left.OculusChromAb2 	= mtl_fx_right.OculusChromAb2 	= floatv(OculusProps.DistortionChromAb[1]); 
	mtl_fx_left.OculusChromAb3 	= mtl_fx_right.OculusChromAb3 	= floatv(OculusProps.DistortionChromAb[2]); 
	mtl_fx_left.OculusChromAb4 	= mtl_fx_right.OculusChromAb4 	= floatv(OculusProps.DistortionChromAb[3]);
	
	camera_left.pan = camera_right.pan = 90;
	vec_set(camera_left.x,  vector(-ipi/2, 0, player_height));
	vec_set(camera_right.x, vector( ipi/2, 0, player_height));
	
	updateFOV();
}



function setDistortionScaleOR(var scale)
{
	OculusProps.DistortionScale = scale;
	mtl_fx_left.OculusScale = mtl_fx_right.OculusScale = floatv(OculusProps.DistortionScale);
	updateFOV();
}

function setEyeToScreenDistanceOR(var esd)
{
	OculusProps.EyeToScreenDistance = esd;
	updateFOV();	
}

function toggleChromaticAberrationOR()
{
	static var enable=0;
	enable=!enable;
	
	effect_load(mtl_fx_left,NULL);
	effect_load(mtl_fx_right,NULL);
		
	if(enable)
	{
		effect_load(mtl_fx_left,"pp_BarrelChromaOR.fx");
		effect_load(mtl_fx_right,"pp_BarrelChromaOR.fx");
	}
	else
	{
		effect_load(mtl_fx_left,"pp_BarrelOR.fx");
		effect_load(mtl_fx_right,"pp_BarrelOR.fx");
	}
}

function toggleDistortionOR()
{
	static var enable=0;
	enable=!enable;
	if(enable)
	{
		camera_left.stage  = camera_left_effect;
		camera_right.stage = camera_right_effect;
	}
	else
	{
		camera_left.stage  = NULL;
		camera_right.stage = NULL;
	}
		
}

function toggleProjectionTranslationOR()
{
	static var enable=0;
	enable=!enable;
	if(enable)
	{
		camera_left.material = mtl_projAdjust_left;
		camera_right.material = mtl_projAdjust_right;
	}
	else
	{
		camera_left.material  = NULL;
		camera_right.material = NULL;
	}
		
}

#endif
